# -*- coding: utf-8 -*-
# @Author: hang.zhang
# @Date:   2017-07-28 11:25:11
# @Last Modified by:   hang.zhang
# @Last Modified time: 2018-03-04 13:51:48

"""
    替换这个的最终目的就是替换引擎和scraper
    为了减少代码量，要上多继承了
    Crawler <- Object
    CrawlerRunner <- Object
    CrawlerProcess <- CrawlerRunner
    关系是 CrawlerRunner 会调用Crawler 从而把三个连起来了
    python 新式类的多继承是：从左到右，宽度优先 (经典类是从左至右，深度优先)
"""


import time
import six
import sys
import logging
from twisted.internet import defer
from scrapy.utils.misc import load_object
from scrapy.crawler import CrawlerProcess, Crawler, CrawlerRunner


logger = logging.getLogger(__name__)

settings_engine_key = "ENGINE"
default_engine = "scrapy.core.engine.ExecutionEngine"


class easyCrawler(Crawler):

    def _create_engine(self):
        """最终目的只是为了替换这个，前面的工作只是为了替换这个
        """
        # 核心变更行 在这里检查settings 是否有写明engine, 否则使用默认的原scrapy engine
        engine = self.settings.get(settings_engine_key, default_engine)
        engine_cls = load_object(engine)
        # return easyExecutionEngine(self, lambda _: self.stop())
        return engine_cls(self, lambda _: self.stop())

    @defer.inlineCallbacks
    def crawl(self, *args, **kwargs):
        """allow spider can return None in start_requests"""
        assert not self.crawling, "Crawling already taking place"
        self.crawling = True

        try:
            self.spider = self._create_spider(*args, **kwargs)
            self.engine = self._create_engine()

            # scrapy's source
            # start_requests = iter(self.spider.start_requests())
            # yield self.engine.open_spider(self.spider, start_requests)

            # modified to:
            start_requests = self.spider.start_requests() or ()
            start_requests = iter(start_requests)
            while True:
                try:
                    # don't close spider in any time !!!!
                    # because sometime adsl server can't not access net, and wrong cause queue empty then let close spider
                    yield self.engine.open_spider(self.spider, start_requests)
                    break
                except Exception:
                    time.sleep(3)
                    logger.exception("error in start spider: self.engine.open_spider")

            yield defer.maybeDeferred(self.engine.start)
        except Exception:
            if six.PY2:
                exc_info = sys.exc_info()

            self.crawling = False
            if self.engine is not None:
                yield self.engine.close()

            if six.PY2:
                six.reraise(*exc_info)
            raise


class easyCrawlerRunner(CrawlerRunner):

    def _create_crawler(self, spidercls):
        if isinstance(spidercls, six.string_types):
            spidercls = self.spider_loader.load(spidercls)
        return easyCrawler(spidercls, self.settings)


class easyCrawlerProcess(CrawlerProcess, easyCrawlerRunner):
    """虽然是多继承，但是实际上CrawlerProcess 和 CrawlerRunner没有重名的方法，定好顺序只是为了以防万一
    """
    pass



# -*- coding: utf-8 -*-
# @Author: hang.zhang
# @Date:   2018-03-05 20:23:52
# @Last Modified by:   hhczy
# @Last Modified time: 2019-12-07 23:17:07
import copy
import json
from scrapy.utils.request import referer_str
from twisted.internet.threads import deferToThread


class commonpipeline(object):
    """通过继承 commonpipeline 来保证

    1. pipeline 是异步同时进行的
    2. 一旦出错，能及时汇报上去，汇报到 failed_urls

    注意, commonpipeline 调用的都是 _process_item  前面有一个 _ 下划线
    """

    def process_item(self, item, spider, response):
        # return self._process_item(item, spider, response)
        # d = deferToThread(self._process_item, item, spider, response)
        # 一定要copy 否则数量会不对，执行的快了，item会相互影响
        copy_item = copy.deepcopy(item)
        # self._process_item(copy_item, spider, response)
        # return copy_item
        d = deferToThread(self._process_item, copy_item, spider, response)

        def error_back(err):
            # 既然出错，那么就要block_callback记录，重新放回起始队列
            r_copy = response.copy()
            # 如果直接从 yiled Reuest 过来的话，那就是没有带上 easyspider 信息的
            if "easyspider" not in r_copy.request.meta:
                r_copy.request.meta["easyspider"] = {"from_retry": 1}
            else:
                r_copy.request.meta["easyspider"]["from_retry"] += 1
            # 带上item 方便检查错误在哪
            # msg = "Error processing %s ; %s" % (json.dumps(item), err.getTraceback())
            msg = {
                "request": "%(request)s (referer: %(referer)s)" % {'request': response.request, 'referer': referer_str(response.request)},
                "body": spider.get_unicode_response_body(response),
                "traceback": "Error processing %s ; %s" % (json.dumps(copy_item), err.getTraceback())
            }
            spider.put_back_2_start_url(r_copy,
                                        exc_info=msg,
                                        )
            return copy_item
        d.addErrback(error_back)
        return d



# -*- coding: utf-8 -*-
# @Author: hang.zhang
# @Date:   2017-07-19 12:00:42
# @Last Modified by:   hang.zhang
# @Last Modified time: 2017-07-19 15:25:33

"""
日志监控,可以重载参数以指定变更内容写入到哪里


CREATE TABLE `easyspider_log` (
  `id` bigint(20) unsigned NOT NULL AUTO_INCREMENT,
  `spider` varchar(100) DEFAULT NULL,
  `logging_time` timestamp NOT NULL DEFAULT CURRENT_TIMESTAMP ON UPDATE CURRENT_TIMESTAMP,
  `logger` varchar(100) DEFAULT NULL,
  `logging_level` varchar(20) DEFAULT NULL,
  `info` text,
  `create_time` timestamp NOT NULL DEFAULT CURRENT_TIMESTAMP,
  `update_time` timestamp NOT NULL DEFAULT CURRENT_TIMESTAMP,
  `state` int(11) DEFAULT NULL,
  PRIMARY KEY (`id`),
  KEY `spider` (`spider`)
) ENGINE=InnoDB AUTO_INCREMENT=13 DEFAULT CHARSET=utf8;


"""

from watchdog.observers import Observer
# from watchdog_file_modify_handler import watchdog_file_modify_handler
from watchdog.events import FileSystemEventHandler
import argparse
import json
import re
import os


monitor_log_file = "file_monitor_log.rate"

log_file_re_check = re.compile(".*\.log$")

log_parser_rex = re.compile("(\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}) (\[[\w\.]+\]) (\w+): (.*?)(?=\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2})\n*")

# def log_parser(content):
#     result = log_parser_rex.findall(content, re.S)
#     return result
def log_parser(content):
    result = re.findall(
        "(\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}) (\[[\w\.]+\]) (\w+): (.*?)(?=\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2})", content, re.S)
    return result

#class base_log_monitor(watchdog_file_modify_handler):
class base_log_monitor(FileSystemEventHandler):

    def modified_callback(self, modified_file_name, modified_content):
        if not self.is_target_file_type(modified_file_name):
            print "not target file %s" % modified_file_name
            return
        # print "file %s modified, content is %s \n\n\n re content is %s" % (modified_file_name, modified_content, log_parser(modified_content))
        # log_content = log_parser(modified_content)
        # if not log_content:
        #     print "no log parse content is %s" % modified_content
        #     return
        
        # print "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!"
        # with open("../xxx.test", "a") as f:
        #     for log in log_content:
        #         logging_time, logger, logging_level, info = log
        #         f.write("%s %s %s %s" % (logging_time, logger, logging_level, info))
        print "is target file %s" % modified_file_name
        with open("../xxx.test", "a") as f:
            f.write(modified_content)

    def is_target_file_type(self, file_name):
        return log_file_re_check.findall(file_name)

    def parse_args(self):
        parser = argparse.ArgumentParser(description="日志监控服务")
        parser.add_argument("-d", help="监控目录，默认为当前目录", default="./")
        args = parser.parse_args()

        return args

    def start_monitor(self):
        args = self.parse_args()
        monitor_path = args.d
        observer = Observer()
        print monitor_path
        observer.schedule(self, path=monitor_path, recursive=True)
        observer.start()
        while True:
            pass
        observer.join()

    def on_modified(self, events):
        modified_file = str(events.src_path)
        if os.path.abspath(modified_file) == os.path.abspath(monitor_log_file) or os.path.isdir(modified_file):
            #print "SKIP, %s" % modified_file
            return
        #print "type is %s %s" % (type(modified_file), type(monitor_log_file))
        #print "moified file name %s, %s == %s is %s" % (modified_file, modified_file, monitor_log_file, os.path.abspath(modified_file)==os.path.abspath(monitor_log_file))
        modified_content = get_last_modify_content(modified_file)
        self.modified_callback(modified_file, modified_content)



def record_file_tail(f, monitor_file_name, record_file=monitor_log_file):
    with open(record_file, "w") as recorder:
        recorder.write(json.dumps(
            {monitor_file_name: f.tell()}, ensure_ascii=False, indent=2))


def resume_last_node(file_name, monitor_file_name=monitor_log_file):
    if not os.path.exists(monitor_file_name):
        print "no monitor_file_name file"
        return 0
    with open(monitor_file_name) as f:
        try:
            monitor_log = json.load(f)
        except Exception, e:
            print e,"file_name %s" % file_name
            return 0
    last_node = monitor_log.get(file_name, 0)
    print "monitor_log.get(%s, 0)  = %s" % (file_name, last_node)
    return last_node


def get_last_modify_content(modified_file):
    last_node = resume_last_node(modified_file)
    modify_content = ""
    with open(modified_file) as f:
        print "seek %s" % last_node
        f.seek(last_node)
        modify_content = f.read()
        record_file_tail(f, modified_file)
    return modify_content




if __name__ == '__main__':
    base_log_monitor().start_monitor()


# -*- coding: utf-8 -*-
# @Author: hang.zhang
# @Date:   2018-03-05 19:20:28
# @Last Modified by:   hhczy
# @Last Modified time: 2019-05-21 14:48:23
from easyspider.pipelines.commonpipeline import commonpipeline
from easyspider.utils.tools import get_time
import socket

flat = (lambda L: sum(map(flat, L), []) if isinstance(L, list) or isinstance(L, tuple) else [L])

import os

class commonBasepipeline(commonpipeline):

    def _process_item(self, item, spider, response):
        item["crawled_time"] = get_time()
        item["spider"] = spider.name
        # os.system("echo %s >> asdhhhh.txt" % item["Ustime"])
        # 多机器抓取用来标志是来源于哪台机器
        try:
            item["crawled_server"] = ";".join(flat(socket.gethostbyname_ex(socket.gethostname())))
        except Exception:
            # 某些情况下可能会出错
            item["crawled_server"] = socket.gethostname()
        # item["crawled_url"] = response.url
        return item



# coding=utf-8
import time
import socket


def parse_time(timestamp):
    try:
        return time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(timestamp))
    except Exception:
        return "0000-00-00 00:00:00"


def get_time():
    return parse_time(time.time())


# 获取IP用
flat = (lambda L: sum(map(flat, L), []) if isinstance(L, list) or isinstance(L, tuple) else [L])


def get_hostname():
    try:
        return ";".join(flat(socket.gethostbyname_ex(socket.gethostname())))
    except Exception:
        # 某些情况下可能会出错
        return socket.gethostname()


if __name__ == '__main__':
    from future import print_function
    print(get_time())



# -*- coding: utf-8 -*-
# @Author: hang.zhang
# @Date:   2018-03-07 23:33:18
# @Last Modified by:   hang.zhang
# @Last Modified time: 2018-03-08 16:06:42
from __future__ import absolute_import
import warnings
import traceback
from scrapy.utils.misc import walk_modules
from scrapy.spiderloader import SpiderLoader
import sys

class easySpiderLoader(SpiderLoader):
    def _load_all_spiders(self):
        for name in self.spider_modules:
            try:
                for module in walk_modules(name):
                    # print module
                    if sys.version_info > (3, ):
                        from importlib import reload
                    else:
                        from imp import reload
                    reload(module)
                    self._load_spiders(module)
            except ImportError:
                if self.warn_only:
                    msg = ("\n{tb}Could not load spiders from module '{modname}'. "
                           "See above traceback for details.".format(
                               modname=name, tb=traceback.format_exc()))
                    warnings.warn(msg, RuntimeWarning)
                else:
                    raise
        try:
            # 1.4 之前的版本，不存在这个函数
            self._check_name_duplicates()
        except AttributeError:
            pass



# coding=utf-8
import logging
from scrapy.exceptions import IgnoreRequest

logger = logging.getLogger(__name__)


class spiderIsAnti(IgnoreRequest):

    def __init__(self, response, *args, **kwargs):
        pass


class antiSpiderMiddleware(object):

    def process_spider_input(self, response, spider):
        # 如果在spider中没有定义isAntiSpider，检测是否被屏蔽的操作，那么就默认都不被屏蔽
        if not hasattr(spider, "is_blocked_spider"):
            return
        # 如果定义了isAntiSpider，那么一般情况下返回False,因为名字是isAntiSpider是否被触发，所以返回False就是没有被触发，返回False就是被触发
        if not spider.is_blocked_spider(response):
            return
        # 如果定义了触发反爬之后的动作，那么就调用这个动作，否则什么都不干
        if hasattr(spider, "blocked_call_back"):
            spider.blocked_call_back(response)
        # 在上面起始已经处理了被屏蔽之后的动作，这里只是记录一下
        raise spiderIsAnti(response)

    def process_spider_exception(self, response, exception, spider):
        if isinstance(exception, spiderIsAnti):
            logger.info(
                "Ignoring response %(response)r: Spider is blocked or not allowed",
                {'response': response}, extra={'spider': spider},
            )
            """
                如果返回为None的话，那么其他中间件的process_spider_exceptions会一直被调用，直到所有中间件被调用，该异常到达引擎(异常将被记录或忽略)
                如果返回一个包含response,dict,item的可迭代对象，那么这个中间件的process_spider_output方法会被调用，其他的process_spider_exception　不会被调用
                第二个返回，就相当于正常调用process_spider_output方法
            """
            return []



# -*- coding: utf-8 -*-
# @Author: hang.zhang
# @Date:   2018-04-04 17:52:25
# @Last Modified by:   hang.zhang
# @Last Modified time: 2018-04-04 17:54:10

from scrapy.statscollectors import MemoryStatsCollector


class easyspiderStatsCollector(MemoryStatsCollector):

    def __init__(self, crawler):

        super(easyspiderStatsCollector, self).__init__(crawler)



# coding=utf-8
from easyspider.spiders.easyCrawlSpider import easyCrawlSpider
from scrapy import Request


class test_redis_cli(easyCrawlSpider):
    
    name = 'baiduSpider_2'

    def start_requests(self):
        for i in xrange(1, 6283 + 1):
            yield Request("http://www.baidu.com/s?wd=%s" % i)

    def parse(self, response):
        print "crawled %s" % response.url



# coding=utf-8

import logging
from six.moves.urllib.parse import urljoin
from scrapy.downloadermiddlewares.redirect import BaseRedirectMiddleware

logger = logging.getLogger(__name__)

"""
    目的：让无限多次的302跳转，不会把请求内容丢弃，【还是返回给spider】让spider的【is_anti_spider】来处理

    1. 只是在原来基础上 _redirect　多加了一个response的参数，
    注释了raise IgnoreRequest("max redirections reached") 加了一个return response
    2. 对于循环请求http://www.baidu.com/saa?wd=%s" % i 这个例子来说，由于baidu把出错的全部导向到error页面
    循环请求的时候，第一次之后，其他由于一样的请求，都被filter了，所以在302的时候，应该不管什么情况都要dont_filter
"""


class myBaseRedirectMiddleware(BaseRedirectMiddleware):

    def _redirect(self, redirected, request, spider, reason, response):
        ttl = request.meta.setdefault('redirect_ttl', self.max_redirect_times)
        redirects = request.meta.get('redirect_times', 0) + 1

        if ttl and redirects <= self.max_redirect_times:
            redirected.meta['redirect_times'] = redirects
            redirected.meta['redirect_ttl'] = ttl - 1
            redirected.meta['redirect_urls'] = request.meta.get('redirect_urls', []) + \
                [request.url]
            # 为什么scrapy源码会写出这一行来 ? 这不是开玩笑吗 ？
            # 你request.replace就复制了request还需要做这个操作？坚决注释替换
            # redirected.dont_filter = request.dont_filter
            redirected.priority = request.priority + self.priority_adjust
            logger.debug("Redirecting (%(reason)s) to %(redirected)s from %(request)s",
                         {'reason': reason, 'redirected': redirected, 'request': request})
            return redirected
        else:
            # 原来的操作是raise 错误，现在是提示，并且放回起始队列
            # logger.info("Discarding %(request)s: max redirections reached",
            #             {'request': request}, extra={'spider': spider})
            # raise IgnoreRequest("max redirections reached")
            # in this time, response do not bind with a request
            # AttributeError: 'NoneType' object has no attribute 'url'
            # response.request = copy.deepcopy(request)
            response.request = request.copy()
            spider.report_this_crawl_2_log(response, "Discarding %(request)s: max redirections reached" % {'request': request})
            # spider.blocked_call_back(response, "Discarding %(request)s: max redirections reached" % {'request': request})
            return response


class directReturnRedirectMiddleware(myBaseRedirectMiddleware):
    """Handle redirection of requests based on response status and meta-refresh html tag"""

    def process_response(self, request, response, spider):
        if request.meta.get('dont_redirect', False):
            return response
        if request.method == 'HEAD':
            if response.status in [301, 302, 303, 307] and 'Location' in response.headers.to_unicode_dict():
                redirected_url = urljoin(
                    request.url, response.headers.to_unicode_dict()['location'])
                redirected = request.replace(url=redirected_url)
                # 所有的跳转都应该被 dont_filter 以防因为屏蔽而被跳转到同一个链接而不能被记录
                # redirected = redirected.replace(dont_filter=True)
                return self._redirect(redirected, request, spider, response.status, response)
            else:
                return response

        if response.status in [302, 303] and 'Location' in response.headers.to_unicode_dict():
            redirected_url = urljoin(request.url, response.headers.to_unicode_dict()['location'])
            redirected = self._redirect_request_using_get(
                request, redirected_url)
            # 所有的跳转都应该被 dont_filter 以防因为屏蔽而被跳转到同一个链接而不能被记录
            # redirected = redirected.replace(dont_filter=True)
            return self._redirect(redirected, request, spider, response.status, response)

        if response.status in [301, 307] and 'Location' in response.headers.to_unicode_dict():
            redirected_url = urljoin(request.url, response.headers.to_unicode_dict()['location'])
            redirected = request.replace(url=redirected_url)
            # 所有的跳转都应该被 dont_filter 以防因为屏蔽而被跳转到同一个链接而不能被记录
            # redirected = redirected.replace(dont_filter=True)
            return self._redirect(redirected, request, spider, response.status, response)

        return response



# coding=utf-8

"""
继承原有的进行修改减少代码,原来默认是报告，现在添加进 redis记录
"""
from scrapy.spidermiddlewares.httperror import HttpErrorMiddleware

from scrapy.utils.serialize import ScrapyJSONEncoder
from scrapy.exceptions import IgnoreRequest
from scrapy_redis import connection
from datetime import datetime
import logging

logger = logging.getLogger(__name__)
default_serialize = ScrapyJSONEncoder().encode


class HttpError(IgnoreRequest):
    """A non-200 response was filtered"""

    def __init__(self, response, *args, **kwargs):
        self.response = response
        super(HttpError, self).__init__(*args, **kwargs)


class recordHttpErrorMiddleware(HttpErrorMiddleware):

    def __init__(self, settings):
        super(recordHttpErrorMiddleware, self).__init__(settings)
        # 与原来相比新增
        self.serialize = default_serialize
        self.server = connection.from_settings(settings)

    def process_spider_exception(self, response, exception, spider):
        if isinstance(exception, HttpError):
            logger.debug(
                "Ignoring response %(response)r: HTTP status code is not handled or not allowed",
                {'response': response}, extra={'spider': spider},
            )
            # 与原来相比新增
            self.recordFailed(response, spider)
            return []

    def recordFailed(self, response, spider):
        # print "\n\n\nin record file\n\n\n"
        source_start_url = response.meta.get("source_start_url")
        if not source_start_url:
            source_start_url = {
                "url": response.meta.get("redirect_urls")[0] if response.meta.get("redirect_urls") else response.url
            }
        source_start_url.update({
            "now_url": response.url,
            "status": response.status,
            "spider": spider.name,
            "time": datetime.utcnow(),
        })
        logger.info(
            "response %(response)r: HTTP status code is not handled or not allowed, the detail msg are %(detail)s",
            {'response': response, "detail": source_start_url}, extra={'spider': spider},
        )
        data = self.serialize(source_start_url)
        self.server.rpush("%s:failed_urls" % (spider.name), data)



# coding=utf-8

# 测试: scrapy crawl spider_name -s SCHEDULER_FLUSH_ON_START=True
# 部署: scrapy crawl spider_name -s LOG_FILE=spider_name.log -s
# LOG_LEVEL=INFO -s LOG_STDOUT=True

REDIS_URL = "redis://127.0.0.1:6379"
MONGO_URL = "mongodb://127.0.0.1:27017/"

MONGO_DB_NAME = "easyspider"

ITEM_PIPELINES = {
    'easyspider.pipelines.pipelines.ExamplePipeline': 500,
    "easyspider.pipelines.easy_mongo_pipeline.easyMongoPipeline": 504,
}

DOWNLOAD_TIMEOUT = 10

# 0的速度太恐怖了
DOWNLOAD_DELAY = 2

# 禁用cookies  因为默认的居然还是开...
COOKIES_ENABLED = False

# 开启后可以看到每次的cookies状态
COOKIES_DEBUG = True

REDIRECT_MAX_TIMES = 3  # 最大重定向次数

# HTTPERROR_ALLOWED_CODES = [200, 301, 302, 303, 304, 305, 306]
# 忽略该列表中所有非200状态码的response
HTTPERROR_ALLOWED_CODES = [301, 302, 303, 304, 305, 306]

DOWNLOADER_MIDDLEWARES = {
    # 修改UA头
    "scrapy.downloadermiddlewares.useragent.UserAgentMiddleware": None,
    "easyspider.middlewares.useragent.userAgentMiddleware": 400,
    # 超过次数的跳转直接返回，不丢弃.跳转全部都是dont_filter
    "scrapy.downloadermiddlewares.redirect.RedirectMiddleware": None,
    "easyspider.middlewares.redirect.directReturnRedirectMiddleware": 600,
    # 连接超时(代理无效)或者50X的错误(其实还有400和408)，就需要重试，超过重试次数，不丢弃give up,而是重新放回入口链接
    "scrapy.downloadermiddlewares.retry.RetryMiddleware": None,
    "easyspider.middlewares.retry.retryToStartMiddleware": 500,
}

SPIDER_MIDDLEWARES = {
    # 类似于404的错误记录到redis中去
    "scrapy.spidermiddlewares.httperror.HttpErrorMiddleware": None,
    "easyspider.middlewares.httperror.recordHttpErrorMiddleware": 50,
    # 是否被block的判断
    "easyspider.middlewares.antiSpider.antiSpiderMiddleware": 950,
}

# 没有不行, 这是scrapy寻找spider的路径
SPIDER_MODULES = ['easyspider.spiders']

# NEWSPIDER_MODULE = 'easyspider.spiders'

# redis 作为去重器
DUPEFILTER_CLASS = "scrapy_redis.dupefilter.RFPDupeFilter"

# 使用自定义的调度器
SCHEDULER = "easyspider.core.scheduler.easyScheduler"

# 显示所有被过滤器过滤掉的请求
DUPEFILTER_DEBUG = True

# 状态收集器scrapy.extensions.logstats打印日志的频率
LOGSTATS_INTERVAL = 1

# 允许自定义引擎
ENGINE = "easyspider.core.easyEngine.easyEngine"

# 允许自定义scraper
SCRAPER = "easyspider.core.easyScraper.easyScraper"



# -*- coding: utf-8 -*-
# @Author: hang.zhang
# @Date:   2017-07-18 16:11:45
# @Last Modified by:   hang.zhang
# @Last Modified time: 2017-08-01 23:34:01
"""
1. 因为比如获取mongo当中结果的操作get_mongo_result不仅仅在mongo_etl到redis中需要被使用
在etl到mysql的时候也需要被使用..这时候要引入一个mongo_etl_redis就不太好，所以在这里就提供
一个基础的etl_base来解决这个问题

2. 同时对于每个etl都有个cmd_parse的函数，实在是太繁杂了，应该是公共的解析要集合在一起才行

3. 同时有时会需要用到某个函数，但是因为函数在类里面，使用必须要初始化，但是初始化又逃不掉
命令行参数的赋予，所以导致很麻烦。这里就干脆除非run的时候去解析命令行参数，其他时候就是可以
单独传参数被调用的。

4. 这个传参数建议不是self的参数，因为是外部调用，只希望使用这个函数，但是不想麻烦的去设置self
的内容
"""
import sys
reload(sys)
sys.setdefaultencoding("utf-8")
sys.path.append("../easyspider/utils")
sys.path.append("../easyspider")
#import settings
#from dbConn_for_etl import MongoConnect, BaseRedis
import argparse
import textwrap
#from logger_base import *


class base_etl(object):

    def __init__(self):
        # mongo系
        self.mongo_server = None
        self.mogno_url = None
        self.mongo_db_name = None

        # redis系
        self.redis_server = None

    # 所有etl的命令行解析函数，参数都不强制要求提供，但是在运行时发现参数不够，报错提示
    # 每个etl的功能，在运行了父类这个函数之后，通过加载-h参数提供
    def parse_cmd(self, desc="etl的基础类,在这里将会解析命令行的参数,参数都不强制要求提供,但是在运行时发现参数不够的时候,将会报错提示"):
        parser = argparse.ArgumentParser(description=desc)

        #------------mongo类-------------------
        parser.add_argument("-mongo-url", help="mongodb连接地址, 默认会采用settings中配置的信息",
                            default=settings.MONGO_URL if hasattr(settings, "MONGO_URL") else "mongodb://127.0.0.1:27017/")
        parser.add_argument("-mongo-db-name", help="mongodb连接的库, 默认会采用settings中配置的信息",
                            default=settings.MONGO_DB_NAME if hasattr(settings, "MONGO_DB_NAME") else "easyspider")
        parser.add_argument("-mongo-collection",
                            help="mongodb的集合, 如果有需要，必须要提供")

        #------------redis类-------------------
        parser.add_argument("-redis-url", help="redis连接地址, 默认会采用settings中配置的信息",
                            default=settings.REDIS_URL if hasattr(settings, "REDIS_URL") else "redis://127.0.0.1:6379")
        parser.add_argument(
            "-redis-key", help="操作的redis-key, 如果同时有mongodb的collection，那么将会以collection名字加上etl前缀")

        #------------csv类---------------------
        parser.add_argument(
            "-csv-file", help="写入的csv文件名, 默认会采用result_csv前缀+collection名")
        parser.add_argument("-csv-split-symbol",
                            help="csv文件的分割符号，默认为 $", default="$")

        #------------mysql类---------------------
        parser.add_argument(
            "-mysql-file", help="写入的mysql脚本文件名, 默认会采用result_mysql_前缀+collection名")
        parser.add_argument(
            "-interval", help="写入mysql脚本时候层级的分割符号，默认为___三个下划线", default="____")
        args = parser.parse_args()
        return args

    #--------------------------mongo系---------------------------------

    # 检查是否已经有mongo连接, 在多次调用函数的时候避免每次重新连接mongodb
    # 返回的是直接可以用的mongo server，直接用mongo server去查询
    # 如果有，检查是否连接地址和库是否都一样，都一样，不用操作
    # 如果有，连接地址一样，数据库地址不一样，那么更新self.mongo_server的库就行了
    # 如果都不对，或者压根没有mongo_server，那么就要重新保存mongo_server
    def mongo_server_check(self, mongo_url, mongo_db_name):
        # 如果已经有了mongo的连接
        if self.mongo_server:
            # 如果连接地址和库都一样，那么就可以使用之前的连接
            if mogno_url == self.mogno_url and mongo_db_name == self.mongo_db_name:
                return self.mongo_server
            # 如果只是连接地址不一样，那么还好可以过来，不需要重新连接
            elif mongo_url == self.mongo_url and mongo_db_name != self.mongo_db_name:
                self.mongo_server.select_db(mongo_db_name)
                return self.mongo_server
        # 如果没有连接，或者连接两个都不对都没有进去return
        self.mongo_server = MongoConnect(mongo_url)
        self.mongo_server.select_db(mongo_db_name)
        # 然后保存最新的mongo的连接地址和连接库
        self.mongo_url, self.mongo_db_name = mongo_url, mongo_db_name
        return self.mongo_server

    # 抽取mongodb中指定collections的全部数据
    def get_mongo_all_result(self, mongo_url, mongo_db_name, collection):
        server = self.mongo_server_check(mongo_url, mongo_db_name)
        return server.find_all(collection)

    # 只抽取部分结果，加快速度方便debug
    def get_mongo_result_limit(self, mongo_url, mongo_db_name, collection, limit=50):
        server = self.mongo_server_check(mongo_url, mongo_db_name)
        return server.find_limit(collection, limit)

    # 把多层的dict平铺，去掉层级关系，层级前缀用_^_^_下划线标志
    # 2017年7月18日12:47:15 心情大好~~ 一遍就过~~~
    # 又加新功能...给定一个过滤函数，在是字典类型的时候开始调用，方便指定不添加哪些字段，不然平铺出来的字段会无线扩充
    def _convert_mongo_2_sheet(self, document, interval, prefix=[], result={}, filter_callback=(lambda x: True)):
        if isinstance(document, dict):
            for k, v in document.items():
                if not filter_callback(k):
                    continue
                prefix.append(k)
                self._convert_mongo_2_sheet(v, interval, prefix, result, filter_callback)
                prefix.pop()
        elif isinstance(document, list):
            # 当list里面放的不是单个的str元素，而是dict又是包含了字典的话就有问题...
            # result[interval.join(prefix)] = interval.join(document)
            # 那么尝试对于list来说，也给他增加一个prefix
            for d in document:
                # 对于list类型来说，不需要上面的prefix
                self._convert_mongo_2_sheet(d, interval, prefix, result, filter_callback)
        else:
            result[interval.join(prefix)] = document
        return result


if __name__ == '__main__':
    bast_etl().parse_cmd()



# -*- coding: utf-8 -*-
# @Author: hang.zhang
# @Date:   2017-07-18 17:23:00
# @Last Modified by:   hang.zhang
# @Last Modified time: 2017-07-18 18:12:07

from base_etl import base_etl
from dbConn_for_etl import BaseRedis
import logging
logger = logging.getLogger(__name__)

"""
python base_mongo_result_2_mysql.py -mongo-collection medicine_all
"""
class base_mongo_start_url_2_redis(base_etl):

    def run(self):
        args = self.parse_cmd()
        
        logger.info("start connect mongo & get result")
        mongo_result = self.get_mongo_all_result(args.mongo_url, args.mongo_db_name, args.mongo_collection)

        # 一般来说拿出来mongo里面的内容要个性化解析之后，才能存到redis里面作为起始链接
        logger.info("get mongo data successed, get %d mongo result, now start clean mongo result" % len(mongo_result))
        clean_result = self.parse_mongo_result(mongo_result)
        
        logger.info("clean mongo result successed, now mongo result count is %d, now start load clean result to redis" % len(clean_result))
        # 如果没有给定redis的名
        if not args.redis_key:
        	args.redis_key = "next_%s:start_urls" % args.mongo_collection
        self.load_clean_result_2_redis(args.redis_url, args.redis_key, clean_result)

        logger.info("all successed !")

    # 这个函数子类可以覆盖，来达到个性化解析mongodb内容的效果
    # 默认情况是，是只提取mongo里面的url字段，并且不做处理直接插入
    # 一种虽然只需要url字段但是需要处理的情况是：url某个字段错了，需要批量替换
    def parse_mongo_result(self, mongo_result):
        clean_result = [{"url": line.get("url")} for line in mongo_result]
        return clean_result

    def load_clean_result_2_redis(self, redis_url, redis_key, clean_result):
        server = BaseRedis(redis_url)
        server.rpush(redis_key, clean_result)



if __name__ == '__main__':
    base_mongo_start_url_2_redis().run()



# -*- coding: utf-8 -*-
# @Author: hang.zhang
# @Date:   2017-07-28 11:07:06
# @Last Modified by:   hhczy
# @Last Modified time: 2019-11-12 22:19:33
"""
    copy from scrapy
"""
import sys
import os
import optparse
import easyspider
from scrapy.cmdline import _run_command
from scrapy.cmdline import _run_print_help
from scrapy.cmdline import _pop_command_name
from scrapy.cmdline import _print_unknown_command
from scrapy.cmdline import _get_commands_from_module
from scrapy.cmdline import _get_commands_from_entry_points
# from scrapy.settings.deprecated import check_deprecated_settings
from scrapy.utils.project import inside_project, get_project_settings

# 目的就是替换掉这个，进而替换掉核心的引擎
# from scrapy.crawler import CrawlerProcess
from easycrawler import easyCrawlerProcess

# 打印出来的提示信息也要变化


def _print_header(settings, inproject):
    if inproject:
        print("easyspider %s - project: %s\n" % (easyspider.__version__,
                                                 settings['BOT_NAME']))
    else:
        print("easyspider %s - no active project\n" % easyspider.__version__)


def _get_commands_dict(settings, inproject):
    cmds = _get_commands_from_module('easyspider.commands', inproject)
    cmds.update(_get_commands_from_entry_points(inproject))
    cmds_module = settings['COMMANDS_MODULE']
    if cmds_module:
        cmds.update(_get_commands_from_module(cmds_module, inproject))
    return cmds


def _print_commands(settings, inproject):
    _print_header(settings, inproject)
    print("Usage:")
    print("  easyspider <command> [options] [args]\n")
    print("Available commands:")
    cmds = _get_commands_dict(settings, inproject)
    for cmdname, cmdclass in sorted(cmds.items()):
        print("  %-13s %s" % (cmdname, cmdclass.short_desc()))
    if not inproject:
        print
        print("  [ more ]      More commands available when run from project directory")
    print
    print('Use "easyspider <command> -h" to see more info about a command')


def execute(argv=None, settings=None):
    if argv is None:
        argv = sys.argv

    # --- backwards compatibility for scrapy.conf.settings singleton ---
    if settings is None and 'scrapy.conf' in sys.modules:
        from scrapy import conf
        if hasattr(conf, 'settings'):
            settings = conf.settings
    # ------------------------------------------------------------------

    if settings is None:
        settings = get_project_settings()
        # set EDITOR from environment if available
        try:
            editor = os.environ.get('EDITOR') or None
        except Exception:
            pass
        else:
            settings['EDITOR'] = editor
    # check_deprecated_settings(settings)
    # ------------------------------------------------------------------

    inproject = inside_project()
    cmds = _get_commands_dict(settings, inproject)
    cmdname = _pop_command_name(argv)
    parser = optparse.OptionParser(formatter=optparse.TitledHelpFormatter(),
                                   conflict_handler='resolve')
    if not cmdname:
        _print_commands(settings, inproject)
        sys.exit(0)
    elif cmdname not in cmds:
        _print_unknown_command(settings, cmdname, inproject)
        sys.exit(2)
    cmd = cmds[cmdname]
    parser.usage = "scrapy %s %s" % (cmdname, cmd.syntax())
    parser.description = cmd.long_desc()
    settings.setdict(cmd.default_settings, priority='command')
    cmd.settings = settings
    cmd.add_options(parser)
    opts, args = parser.parse_args(args=argv[1:])
    _run_print_help(parser, cmd.process_options, args, opts)

    # 实际上只变更了这么一行
    # cmd.crawler_process = CrawlerProcess(settings)
    cmd.crawler_process = easyCrawlerProcess(settings)
    _run_print_help(parser, _run_command, cmd, args, opts)
    sys.exit(cmd.exitcode)


if __name__ == '__main__':
    execute()



# -*- coding: utf-8 -*-
# @Author: hang.zhang
# @Date:   2017-09-05 16:20:40
# @Last Modified by:   hang.zhang
# @Last Modified time: 2018-03-04 14:02:54

from scrapy_redis.dupefilter import RFPDupeFilter
import logging


logger = logging.getLogger(__name__)


class easyRFPDupeFilter(RFPDupeFilter):

    def request_seen(self, request):
        """
        1. 增加try except 避免在机器断网的时候，无法连接redis避免崩溃结束
        2. 改变 dupefilter插入时间，只有当request能产生结果，才会被加入dupefilter
        (即是否重复的标准，是否请求过，变成是否产生结果)。

        主要是应对以下场景： 如果请求被屏蔽，那么这个请求，就无法获得结果。

        而根据原始的逻辑来看，这个请求你确实已经发出了，并且有响应。那么你这个请求在发出的时候，就已经加入了dupefilter，下次就不会再被调度，【请求就会被丢失】

        这样做的目的，是精细化抓取，严格要求，不丢掉一条请求
        """
        fp = self.request_fingerprint(request)
        # This returns the number of values added, zero if already exists.
        try:
            # added = self.server.sadd(self.key, fp)
            added = self.server.sismember(self.key, fp)
            return added
        except Exception:
            # return false to schedler this request
            logger.exception("in easyRFPDupeFilter: check is in dupefilter failed, return false")
            return False



# -*- coding: utf-8 -*-
# @Author: hang.zhang
# @Date:   2018-03-05 20:20:45
# @Last Modified by:   hang.zhang
# @Last Modified time: 2018-04-03 14:13:41

# 历史负担真是重啊....为什么说 unexcept argument...原来是引用错包了
# from easyspider.utils.DBService import MongoService
from DBService import MongoService
from easyspider.pipelines.commonpipeline import commonpipeline
import logging
import time
logger = logging.getLogger(__name__)

# 存入mongo的collection的key名
item_result_table_key = "result_table"
# 存入的操作是update还是直接insert
item_update_key = "update_record"
item_update_query_key = "update_query_key"
# 存入的时候是否需要新建一个collections来保存历史记录
item_keep_history = "keep_history"
# 存入库的名字，如果没有指定库，那么就以spider名字来命名
item_save_db = "save_db"

default_mongo_url = "mongodb://localhost:27017"
default_mongo_db_name = "spider"

mongo_url_key = "MONGO_URL"
mongo_db_name = "MONGO_DB_NAME"

# ----------------------------------------
# 上面这一段，是历史原因保留
# ----------------------------------------


class commonMongopipeline(commonpipeline):

    def __init__(self, mongoUrl, mongoDbName):
        self.server = MongoService(mongoUrl)
        self.server.select_db(mongoDbName)

    @classmethod
    def from_settings(cls, settings):
        mongoUrl = settings.get(mongo_url_key, default_mongo_url)
        mongoDbName = settings.get(mongo_db_name, default_mongo_db_name)
        return cls(mongoUrl, mongoDbName)

    @classmethod
    def from_crawler(cls, crawler):
        return cls.from_settings(crawler.settings)

    def _process_item(self, item, spider, response, db=None, collection=None, update=False, hash_check_item=None, hash_check_item_dict_type=False):

        # 解析返回结果里面的控制状态,注意是有两层
        easyspider_meta = item.get("easyspider", {}).get("mongo_config", {})
        if not easyspider_meta:
            # 兼容以前的老的写法
            easyspider_meta = item.get("easyspider", {}).get("save_mongo", {})
        # print "\n\n\n", easyspider_meta, "\n\n"

        # 解析控制状态
        db = easyspider_meta.get("db")
        collection = easyspider_meta.get("collection")
        # 是否打开insert_or_update
        update = easyspider_meta.get("update")
        hash_check_item = easyspider_meta.get("hash_check_item")
        hash_check_item_dict_type = easyspider_meta.get("hash_check_item_dict_type")
        force_replace = easyspider_meta.get("force_replace")

        error_flag = False
        error_limit = 3
        error_count = 1
        while error_count <= error_limit:
            try:
                self._insert_or_update(item, spider, db, collection, update, hash_check_item, hash_check_item_dict_type, force_replace)
                error_flag = False
                break
            except Exception:
                error_flag = True
                logger.exception(u"第[%s / %s]操作数据库失败，下次再尝试" % (error_count, error_limit))
                time.sleep(2)

            error_count += 1

        # 如果最后一次还是错了，那么再来一次，还是不行，就算了
        if error_flag:
            self._insert_or_update(item, spider, db, collection, update, hash_check_item, hash_check_item_dict_type, force_replace)

        return item

    def _insert_or_update(self, item, spider, db=None, collection=None, update=False, hash_check_item=None, hash_check_item_dict_type=False, force_replace=None):
        # 避免改变原值
        _save_item = item.copy()

        # 选择指定的库
        if not db:
            self.server.select_db(_save_item.get(item_save_db, spider.name))
        else:
            self.server.select_db(db)

        # print "sleect db %s" % db

        # 选择指定的集合
        if not collection:
            collection = _save_item.get(item_result_table_key, spider.name)

        # 如果不是更新操作的话，那么很简单的直接插入就可以了
        if not update:
            self.server.insert(collection, _save_item)
            return item

        # 更新有两种，第一种是全盘更新，全盘覆盖，也就是 replace
        # 另外一种就是指定列的update 不影响其他的列
        # 默认为部分更新
        if not force_replace:
            # 如果没有指定，或者force_repalce为False
            force_replace = False
        else:
            force_replace = True

        # 如果是更新操作，那么就比较麻烦了

        if hash_check_item_dict_type:
            # 如果更新操作，是给定了一个特定的值，也就是替换的，不是现有的item的特定值
            last_result = hash_check_item
        else:
            # 如果是根据现在给出的结果，进行搜索替换的话
            last_result = {}
            for k in hash_check_item:
                last_result[k] = _save_item.get(k)
        # print "last_result %s , _save_item %s" % (last_result, _save_item)
        # upsert = True 控制没有就是新增

        # 这里要判断，到底是替换，还是部分更新
        if force_replace:
            self.server.replace_one(collection, last_result, _save_item, upsert=True)
        else:
            self.server.update_one(collection, last_result, {"$set": _save_item}, upsert=True)

        # 最后替换操作完成了之后再来返回
        return item



# coding=utf-8
import sys
if sys.version_info < (3, ):
    reload(sys)
    sys.setdefaultencoding("utf-8")
    import urlparse
else:
    from urllib import parse as urlparse
from scrapy_redis.spiders import RedisCrawlSpider
from easyspider.utils.tools import get_time
from easyspider.utils.tools import flat
from scrapy.http import Response
from bs4 import UnicodeDammit
from scrapy import Request
import traceback

import urllib
import socket
import time
import json
import copy
import pdb
import redis


class easyCrawlSpider(RedisCrawlSpider):
    """所有爬虫最全的基类
    提供了一系列优秀简便的方法，来使编码过程更加友好快捷
    base spider of easyspider
    have provied multi useful method to make crawler more easier
    """

    name = "easyCrawlSpider"
    # redis 数据库key的定义，都是按照 spider_name:key_name 的格式
    start_key = "start_urls"
    priority_start_key = "priority_start_urls"
    successed_key = "successed_urls"
    failed_key = "failed_urls"
    # 最大重试次数
    retry_max_time = 1

    # 一次从redis中取多少条任务/链接
    fetch_unit = 16

    # 操作redis 的方法，可替换
    # 操作普通队列的方法， rpop, lpush
    fetch_from_redis_method = (lambda self, server: server.rpop)
    save_into_redis_method = (lambda self, server: server.lpush)

    # 操作优先级队列的方法
    fetch_from_priority_queue = (lambda self, server, key_name, fetch_unit: server.zrevrangebyscore(key_name, float('inf'), float('-inf'), 0, fetch_unit))
    delete_from_priority_queue = (lambda self, server, key_name, val: server.zrem(key_name, val))
    if redis.VERSION[0] < 3:
        save_into_priority_queue = (lambda self, server, key_name, val, data: server.zadd(key_name, val, data))
    else:
        save_into_priority_queue = (lambda self, server, key_name, val, data: server.zadd(key_name, {data: val}))

    def next_requests(self):
        """
        整个的心跳函数
        heartbeat of engine, called every 5s, to generate new request in crawler process

        next_requests will [r]pop the spider_name:start_key list queue from redis (attention, operation is lpop,
        so when you need to push into redis, you should use lpush instead)
        """
        count = 0
        while count < self.fetch_unit:
            """从redis中获取任务，进行编码，然后提交到引擎，注意：

            1. 先检查 优先队列中是否有任务，优先队列中有，先处理优先队列的任务
            2. 其次再是正常的list普通任务

            (try except 是因为：在整个过程中，由于adsl 机器恶劣的网络环境，每一步网络操作，都需要预防在操作的过程中出现 【断网】的极端情况)
            """
            try:
                # 先尝试从优先队列中获取任务
                start_map = self.fetch_from_priority_queue(self.server, "%s:%s" % (self.name, self.priority_start_key), 1)
                if start_map:
                    """注意返回的是list，所以需要[0]的操作
                    另外优先队列是没有 pop的操作的，get了之后，需要立即做pop/delete处理
                    """
                    start_map = start_map[0]
                    self.delete_from_priority_queue(self.server, "%s:%s" % (self.name, self.priority_start_key), start_map)
                else:
                    """如果优先队列中没有内容，则从普通任务中提取
                    if not, fetch from common queue
                    """
                    start_map = self.fetch_from_redis_method(self.server)(u"%s:%s" % (self.name, self.start_key))
            except Exception:
                """如果因为某种问题，如断网，影响了redis的连接，此时应该等待一段时间
                再次尝试连接。

                因为没有获得具体的任务，所以下面的操作都执行不下去，在这里就要continue，继续尝试从redis中获取任务
                """
                self.logger.exception("in easyCrawlSpider: next_requests,  start_map = self.fetch_from_redis_method(self.server)(u\"%s:%s\" % (self.name, self.start_key)) failed, continue")
                # time.sleep(3)
                count += 1
                continue
            # 提取任务成功，开始对任务进行解码，提交到引擎处理
            req = self.convert_startmap_2_request(start_map)
            if isinstance(req, Request):
                yield req
                # 收到的任务数+1
                count += 1
            else:
                # 如果没有获取到任务，那么就直接退出，等待下次检查是否有新任务出现
                break

        self.logger.debug("Read %s requests from '%s:%s'" % (count, self.name, self.start_key))

    def convert_startmap_2_request(self, start_map):
        """爬虫任务的解析函数。redis 中存储的任务，及任务附带的信息，在这里被解码，从而
        真正提交到引擎去执行抓取。
        注意这个解析是非常复杂的，涉及了非常多的方面。
        """

        # 如果redis中任务为空，那么直接返回，等待下次检查是否有新任务出现
        if not start_map:
            return
        # 如果任务不能被解码，那么就是极度异常的情况，后面都不能继续下去
        try:
            start_map = json.loads(start_map)
        except Exception:
            self.logger.exception(u"convert start url -> dict failed, start_map source is %s" % (start_map))

        try:
            """解析原则：

            1. 如果这个任务，是由 retry_middleware (即是属于，在出错情况下，被添加进来的任务)那么就要无视他的dont_filter(因为本来就是出错才重试，如果被filter了，那么就没有意义)

            """

            # ------------------------------------------
            # 处理dont_filter:
            # 判断是否是，之前任务执行失败，而添加的新任务，就是看任务中，有没有 from_retry字段
            if not start_map.get("dont_filter"):
                # 含有from_retry 字段，且from_retry 不为0
                # from_retry = int(start_map.get("easyspider", {}).get("from_retry", 0))
                # if from_retry > 0:
                #     close_filter = True
                # else:
                #     close_filter = False
                close_filter = False
            else:
                # 其他情况，那么就根据他自己的情况来, 指定关闭过滤器，则关闭过滤器
                if start_map.get("dont_filter") in ["True", "true", True]:
                    close_filter = True
                else:
                    close_filter = False

            # ------------------------------------------
            # 处理callback:
            # callback 函数默认为self.parse, 但是一样可以自定义
            # 不过要注意的是，需要把callback由字符串，转变为函数

            callback_fun = start_map.get("callback") or "self.parse"
            req_callback_method = eval(callback_fun)

            # ------------------------------------------
            # 处理url:
            # 请求的网址，从url 中获取

            req_url = start_map.get("url")

            # ------------------------------------------
            # 处理请求方法:
            # 默认的请求是 GET

            req_method = start_map.get("method") or "GET"

            # ------------------------------------------
            # 处理请求头:
            # 注意，请求头应该是dict字典格式，而不是编码后的字符串 ！！！
            # 需要额外注意的是，如果是json格式的请求，需要额外编码

            req_headers = start_map.get("headers") or {}

            # ------------------------------------------
            # 处理cookie：
            # 注意，cookie 应该是字符串格式，这里将不会再进行编码

            req_cookies = start_map.get("cookies") or None

            # ------------------------------------------
            # 处理请求体body:
            # 请求体可以有多种格式，无论是dict字典的给出，还是字符串的给出，都能被接受
            # 注意，针对json格式的请求，需要额外增加支持

            if req_method.upper() != "GET":
                post_data = start_map.get("body", "")
                req_body = post_data
                req_body = post_data.encode("utf-8")
                # if isinstance(post_data, dict):
                #     req_body = urllib.urlencode(post_data)
                #     # req_body = post_data
                # elif isinstance(post_data, unicode) or isinstance(post_data, str):
                #     req_body = post_data

                # 针对json 格式的请求，增加支持
                # if "application/json" in req_headers.get("Content-Type", ""):
                #     req_body = json.loads(req_body)
            else:
                req_body = None
            # print req_body
            # pdb.set_trace()
            # ------------------------------------------
            # 处理其他的特色附加信息，这个也是非常重要的
            # 特色附加信息，都是挂载在 easyspider 下的
            # 注意这里忽略了 exo_info 错误信息
            # 2018年03月04日18:07:51   以下都是错的，作为一个解析任务的函数，为什么还需要做处理呢
            # 直接带上就好了啊
            req_meta = start_map.get("meta", {}) or {}
            # last_easyspider = req_meta.get("easyspider") or {}
            # easyspider_info = {
            #     # 是否来源于错误提交
            #     "from_retry": last_easyspider.get("from_retry", 0),
            #     # 真实的起始链接，如果没有，就是当前的链接
            #     "source_start_url": last_easyspider.get("source_start_url") or req_url,
            #     # 备注信息，一般没有
            #     "remark": last_easyspider.get("remark"),
            #     # 爬虫经过的全部路径，调试的非常重要，这样就知道为什么被卡住了
            #     "crawled_urls_path": last_easyspider.get("crawled_urls_path") or []
            # }
            # req_meta.update({"easyspider": easyspider_info})

            # ------------------------------------------
            # 最终才能够组合完成
            return Request(url=req_url,
                           callback=req_callback_method,
                           method=req_method,
                           meta=req_meta,
                           body=req_body,
                           headers=req_headers,
                           cookies=req_cookies,
                           dont_filter=close_filter
                           )
        except Exception:
            self.logger.exception(u"parse start url error, source start url is %s" % start_map)

    def is_blocked_spider(self, response):
        """检测爬虫是否被屏蔽的方法。每个成功的请求，都会进入这个函数
        默认为 False, 即爬虫没有被屏蔽，没有被屏蔽，那么就会进入正常流程，接下来调用对应callback

        可以重写这个函数，来检测爬虫是否被屏蔽，比如检测
        u"次数过多" in resposne.body
        出现这个情况的时候，就是True，爬虫就被屏蔽，就不会进入接下来的callback正常流程
        接下里就会调用blocked_call_back， 来进行屏蔽处理

        method is called every time to check if blocked the spider, default to be False or None
        """
        pass

    def blocked_call_back(self, response, reason="spider was blocked", exc_info=None, extra=None):
        """兼容历史老代码的api
        """
        self.blocked_callback(response, reason, exc_info, extra)

    def blocked_callback(self, response, reason=u"spider was blocked", exc_info=None, extra=None):
        """如果检测到爬虫被屏蔽，那么就会调用这个函数来对进行处理
        默认的操作是重新放回 任务队列的队尾，可以通过重写这个方法，来完成自定义的 屏蔽处理逻辑

        注意一点：这个函数，并不只是在被屏蔽的时候可以调用，它是 【通用的】
        即，当你需要重新抓取某个链接的时候，可以通过 【假装这个链接被屏蔽了】，来达到重新抓取的效果

        一个典型应用就是： 我在列表页list发现了新的任务，通过调用blocked_call_back 的put_back_2_start_url, 达到添加任务的效果

        method called when spider is blocked, also can be used to record other error situation"""

        # 记录这个状态，即爬虫出现问题
        self.report_this_crawl_2_log(response, reason, exc_info)
        # 重新添加任务
        if not exc_info:
            # 如果没有exc_info 那么说明就是从 is_spider_blocked 过来的，附带上被屏蔽返回的代码
            exc_info = "%s; response.body %s" % (reason, self.get_unicode_response_body(response))
        self.put_back_2_start_url(response, exc_info)

    def report_this_crawl_2_log(self, response, reason, exc_info=None):
        """记录此次请求
        log this response
        """
        if not exc_info:
            # 在 redirectmiddleware 中，是直接的reason，没有返回exc_info的
            # 适应之前代码
            exc_info = reason
        report_template = u"""
            response body -> %(response_body)s
            %(response)s is recorded, because %(reason)s happended, following are detail info:
            response url -> %(response_url)s,
            status code -> %(status_code)s,
            request url -> %(request_url)s,
            original request url -> %(original_request_url)s

            request headers -> %(request_headers)s,
            request body -> %(request_body)s,

            request callback -> %(request_callback)s,

            exc_info -> %(exc_info)s
            """

        report_info = {
            "response": response.url,
            "reason": reason,
            "response_url": response.url,
            "status_code": response.status,
            "request_url": response.request.url,
            "original_request_url": self.get_source_url(response),
            "request_headers": response.request.headers.to_unicode_dict(),
            # dict type
            "request_body": self.get_request_body(response),
            # u"request_callback": response.request.callback.__name__,
            "request_callback": self.get_last_request_callback(response),
            # unciode body
            "response_body": self.get_unicode_response_body(response),
            "exc_info": exc_info,
        }

        self.logger.warning(report_template % report_info)

    def get_common_request(self, r):
        """兼容Reques和Response 的状态，统一变为request状态
        避免修改，返回的都是副本

        注意：Request object can't use copy
        """
        if isinstance(r, Response):
            # Request object can't use copy
            # r_copy = copy.deepcopy(r.request)
            r_copy = r.request.copy()
        else:
            # r_copy = copy.deepcopy(r)
            r_copy = r.copy()
        return r_copy

    def get_source_url(self, r):
        """获得最原始的的请求链接
        因为在302跳转的情况下，当前的链接，并不是最开始的任务链接
        to get the original url, because 302 will modified the request url
        """
        r_copy = self.get_common_request(r)

        source_url = r_copy.meta.get("easyspider", {}).get("source_start_url")
        redirect_urls = r_copy.meta.get("redirect_urls") or []

        if source_url:
            # 如果source_start_url里面有内容(即在任务里面，就注明了初始url 这个优先级是最高的)
            return source_url
        elif len(redirect_urls) > 0:
            # 第二优先级，是取redirect_urls里面的第一个url
            return redirect_urls[0]
        else:
            # 实在是什么都没有，那么初始链接，就是当前的链接
            return r_copy.url

    def get_request_body(self, r):
        """把字符串的请求体，变成字典类型，方便人类查看
        return str/unicode request body -> dict
        """
        r_copy = self.get_common_request(r)
        return r_copy.body.decode("utf-8")
        dict_body = self.parse_query_2_dict(r_copy.body)

        # pdb.set_trace()
        if not dict_body:
            try:
                # dict_body = json.dumps(r_copy.body, ensure_ascii=False)
                dict_body = r_copy.body
            except Exception:
                self.logger.exception("convert request body failed... source body is %s" % r_copy.body)
        return dict_body

    def parse_query_2_dict(self, query):
        """parse url query to dict format"""
        try:
            return dict([(k, v[0]) for k, v in urlparse.parse_qs(query).items()])
        except Exception:
            self.logger.exception("can't parse url -> dict, source url data is %s" % query)

    def detect_encoding(self, body):
        dammit = UnicodeDammit(body)
        return dammit.original_encoding

    def get_unicode_response_body(self, response):
        """避免乱码，在保存的时候，所有的body 都会被解码成unicode形式
        """
        return response.body_as_unicode()
        encoding = self.detect_encoding(response.body)
        if encoding:
            # 有编码则解码，无编码则直接返回，因为无编码的话，按照None解码将会出错
            # TypeError: decode() argument 1 must be string, not None
            # print u"检测编码结果是：%s" % encoding
            try:
                # 如果请求的是图片，那么检测出来的虽然是utf-8 但是是无法解码的
                # 保存直接保存整个body 就可以了
                return response.body.decode(encoding)
            except Exception:
                msg = traceback.format_exc()
                return u"decode response error, msg is %s" % msg
        return response.body

    def put_back_2_start_url(self, r, exc_info=None, last_response=None):
        """非常有效的核心通用函数
        这个函数不只是在 检测是否被屏蔽的时候可以调用

        还可以被充当，添加新任务的功能。
        如果需要添加一个新的任务，那么直接调用这个函数就可以了

        reput request to redis
        """

        r_copy = self.get_common_request(r)

        # 基于这样一种情况，如果某个response 不是正确的response
        # 那么就根据这个错误response，对应的request，生成新的request
        if not last_response:
            last_response_copy = r_copy
        else:
            last_response_copy = self.get_common_request(last_response)

        # 检查这个任务，是否需要提高优先级，默认优先级为0
        if hasattr(r_copy, "priority"):
            priority = r_copy.priority or 0
        else:
            priority = 0

        # 最原始的reput_request
        reput_request = {
            "url": self.get_source_url(r_copy),
            "callback": self.get_last_request_callback(r_copy),
            "method": r_copy.method,
            "headers": r_copy.headers.to_unicode_dict(),
            "dont_filter": False if (not hasattr(r, "dont_filter") or r.dont_filter is None) else r.dont_filter,
            "priority": priority,
        }

        # ------------------------------------------
        # 检测请求体：
        # 如果这个请求不是GET类型，那么需要添加请求体body
        reput_request["body"] = self.get_request_body(r_copy)
        if reput_request.get("method") == "GET":
            reput_request.pop("body")

        # ------------------------------------------
        # 更新附加信息

        meta = copy.deepcopy(r_copy.meta)
        # pdb.set_trace()
        last_easyspider = meta.get("easyspider") or {}
        last_easyspider.update({
            "from_retry": (int(last_easyspider.get(u"from_retry", -1)) + 1),
            "source_start_url": self.get_source_url(r_copy),
            "remark": last_easyspider.get(u"remark"),
            "crawled_urls_path": last_easyspider.get("crawled_urls_path", []) + [last_response_copy.url],
            "exc_info": exc_info
        })
        # ------------------------------------------
        # 记录当前的时间，和出现问题时候的IP(机器)
        # ------------------------------------------
        try:
            last_easyspider["crawled_server"] = ";".join(flat(socket.gethostbyname_ex(socket.gethostname())))
        except Exception:
            # 没想到这里也引用了
            last_easyspider["crawled_server"] = socket.gethostname()
        last_easyspider["crawled_time"] = get_time()

        meta["easyspider"] = last_easyspider
        reput_request["meta"] = meta

        # ------------------------------------------
        # 检测cookie:
        # 如果开启了cookie，那么cookie随之保存启用
        if self.settings.get(u"COOKIES_ENABLED"):
            reput_request["cookies"] = r_copy.cookies

        # ------------------------------------------
        # 开始存入 redis 任务队列，存入不成功，一直存入
        current_retry_time = reput_request.get("meta").get("easyspider").get("from_retry")
        while True:
            try:
                max_retry_time = self.retry_max_time + 1
                try:
                    dumps_str = json.dumps(reput_request, ensure_ascii=False)
                except Exception:
                    try:
                        traceback.print_exc()
                        import pdb
                        pdb.set_trace()
                        exc_info = json.dumps(reput_request["meta"]["easyspider"]["exc_info"], ensure_ascii=False)
                    #     json.dumps(exc_info, ensure_ascii=False)
                    #     json.dumps(reput_request["meta"]["easyspider"], ensure_ascii=False)
                        reput_request["meta"]["easyspider"]["exc_info"] = exc_info.decode("gbk")
                        # reput_request["meta"]["easyspider"]["exc_info"] = exc_info.decode("gbk")

                        dumps_str = json.dumps(reput_request, ensure_ascii=False)
                    except Exception:
                        #     traceback.print_exc()
                        #     traceback.print_exc()
                        traceback.print_exc()
                        # pdb.set_trace()
                if current_retry_time >= max_retry_time:
                    # 如果超过了重试次数的限制，那么就放回failed
                    try:
                        self.save_into_redis_method(self.server)("%s:%s" % (self.name, self.failed_key), dumps_str)
                    except Exception:
                        exc_info = json.dumps(reput_request["meta"]["easyspider"]["exc_info"], ensure_ascii=False)
                        reput_request["meta"]["easyspider"]["exc_info"] = exc_info
                        # json.dumps(reput_request["meta"]["easyspider"], ensure_ascii=False)
                        try:
                            self.save_into_redis_method(self.server)("%s:%s" % (self.name, self.failed_key), json.dumps(reput_request, ensure_ascii=False))
                        except:
                            traceback.print_exc()
                            # pdb.set_trace()
                # elif reput_request.get("priority") != 0:
                elif reput_request.get("priority") > 0:
                    # 有时候priority还会出现负数的情况
                    # 如果是优先任务，那么放入优先任务队列
                    self.save_into_priority_queue(self.server, "%s:%s" % (self.name, self.priority_start_key), float(reput_request.get("priority")), dumps_str)
                else:
                    # 其他的放入普通任务
                    self.save_into_redis_method(self.server)("%s:%s" % (self.name, self.start_key), dumps_str)
                break
            except Exception:
                msg = traceback.format_exc()
                self.logger.exception("in easyCrawlSpider: put_back_2_start_url,  self.save_into_redis_method failed, loop until it successed.. reput_request is %s, %s" % (reput_request, msg))
                # break
                # current_retry_time += 1
                # time.sleep(2)
                # 需要重新加入，而不是简单的在这里退出，因为涉及到最后对 exc_info 字段的处理，当前进来第一次exc_info是空的，要等到后面自动来添加
                # self.put_back_2_start_url(r)
                # raise Exception(msg)

        # logginig
        put_back_logging_template = """reput crawl task into start url, detail info are %s"""
        self.logger.info(put_back_logging_template % reput_request)

    def get_last_request_callback(self, r):
        r_copy = self.get_common_request(r)
        if not r_copy.callback:
            return "self.parse"
        return "self.%s" % r_copy.callback.__name__

    def parse_request_to_dict(self, r):
        """这家伙不能删除。。。在 core/easyScraper.py 中会调用。。坑爹
        return error request to start step
        """
        if isinstance(r, Response):
            # r_copy = copy.deepcopy(r.request)
            r_copy = r.request.copy()
        else:
            # r_copy = copy.deepcopy(r)
            r_copy = r.copy()

        reput_request = {
            u"url": self.get_source_url(r_copy),
            # u"callback": r_copy.callback.__name__,
            u"callback": self.get_last_request_callback(r_copy),
            u"method": r_copy.method,
            u"body": self.get_request_body(r_copy),
            u"headers": r_copy.headers.to_unicode_dict(),
            u"dont_filter": r_copy.dont_filter
        }

        meta = copy.deepcopy(r.meta)
        # direct launch Requst from start_requests won't have start_url template
        meta.get(u"easyspider", {}).update({
            u"from_retry": int(meta.get(u"easyspider", {}).get(u"from_retry", 0) or 0) + 1,
            u"source_start_url": self.get_source_url(r_copy),
            u"remark": meta.get(u"easyspider", {}).get(u"remark"),
        })
        reput_request["meta"] = meta
        # TODO: check if need to save cookies
        if self.settings.get(u"COOKIES_ENABLED"):
            reput_request["cookies"] = r_copy.cookies
        else:
            reput_request["cookies"] = None
        return reput_request

    def extract_text_and_strip(self, xpath_exp):
        """用来方便的获取某个下面所有的text"""
        return "".join(xpath_exp.xpath(".//text()").extract()).strip()

    def text_strip(self, t):
        """增强型strip 用来方便的过滤掉其他的乱字符"""
        return t.strip().replace("\r", "").replace("\n", "").replace("\t", "")



# -*- coding: utf-8 -*-
# @Author: hang.zhang
# @Date:   2017-07-28 13:54:28
# @Last Modified by:   hang.zhang
# @Last Modified time: 2018-03-04 13:52:32
from scrapy.core.engine import ExecutionEngine
from scrapy.utils.misc import load_object
import logging

logger = logging.getLogger(__name__)

scraper_key = "SCRAPER"
default_scraper = "scrapy.core.scraper.Scraper"


class easyEngine(ExecutionEngine):

    def __init__(self, crawler, spider_closed_callback):
        """ExecutionEngine 在初始化的时候定义了scraper
        这里等他初始化完了,再更改过来了,就无需复制之前的初始化代码
        """
        super(easyEngine, self).__init__(crawler, spider_closed_callback)

        scraper = self.settings.get(scraper_key, default_scraper)
        scraper_cls = load_object(scraper)
        self.scraper = scraper_cls(crawler)



# -*- coding: utf-8 -*-
# @Author: hang.zhang
# @Date:   2017-07-28 14:07:10
# @Last Modified by:   hhczy
# @Last Modified time: 2019-12-07 23:16:45
import sys
if sys.version_info < (3, ):
    reload(sys)
    sys.setdefaultencoding("utf-8")
import os
import json
import logging
import traceback
from scrapy import signals
from scrapy.http import Request
from scrapy.item import BaseItem
from scrapy_redis import picklecompat
from scrapy.exceptions import DropItem
from scrapy.core.scraper import Scraper
from twisted.python.failure import Failure
from scrapy.utils.request import referer_str
from scrapy.utils.reqser import request_to_dict
from scrapy.utils.log import failure_to_exc_info
from scrapy.utils.log import logformatter_adapter
from scrapy.utils.request import request_fingerprint


logger = logging.getLogger(__name__)


class easyScraper(Scraper):

    # 核心希望替换的东西
    def handle_spider_error(self, _failure, request, response, spider):
        """
            目的就是要在出错的时候，能够打印出出错的body请求。同时把出错的请求，重新放回队列中去
            预防这样一种情况：当出现正则或者xpath规则提取不到的时候，能够马上反映出，是为什么规则出错
            是否是发现了不同格式的body内容，需要增加规则的数量
            而常规的scraper 只是记录报错写error log和reason，并不会告诉你此时的response body
            由于blocked之后的操作和此次目的是一样的，所以直接就用这个
        """

        super(easyScraper, self).handle_spider_error(_failure, request, response, spider)
        # traceback_msg = _failure.getErrorMessage()
        traceback_msg = "%s, %s" % (failure_to_exc_info(_failure), repr(traceback.format_exc()))
        if os.name == "nt":
            traceback_msg = traceback_msg.decode("gbk")
        msg = {
            "request": "%(request)s (referer: %(referer)s)" % {'request': request, 'referer': referer_str(request)},
            "body": spider.get_unicode_response_body(response),
            "traceback": traceback_msg
        }
        """
        msg = "Spider error processing %(request)s (referer: %(referer)s): response body %(body)s  traceback %(traceback)s" % {'request': request,
                                                                                                                               "referer": referer_str(request),
                                                                                                                               "body": spider.get_unicode_response_body(response),
                                                                                                                               "traceback": traceback_msg
                                                                                                                               }
        """
        spider.blocked_call_back(response,
                                 reason="Spider error processing %(request)s (referer: %(referer)s): info %(info)s" % {'request': request, 'referer': referer_str(request), 'info': "%s, %s" % (failure_to_exc_info(_failure), repr(traceback.format_exc()))},
                                 exc_info=msg,
                                 extra=self.__class__.__name__)

    # Error downloading 就不需要记录，因为Error downloading根本没有body
    # def _log_download_errors(self, spider_failure, download_failure, request, spider):
    #     pass

    def _process_spidermw_output(self, output, request, response, spider):
        """给pipeline结果添加response参数，这样方便根据response回推
        add response argument in item pipeline
        """
        if isinstance(output, Request):
            meta_easyspider = output.meta.get("easyspider") or {}
            meta_easyspider.update({
                "crawled_urls_path": meta_easyspider.get("crawled_urls_path", []) + [request.url]
            })
            output.meta.update({"easyspider": meta_easyspider})
            self.crawler.engine.crawl(request=output, spider=spider)
        elif isinstance(output, (BaseItem, dict)):
            self.slot.itemproc_size += 1
            # this if will never be false, because the only time to be false, is request from put_back_2_start_urls, and history request has meta..and has
            # crawled_urls_path...but, ! put_back_2_start_urls will not record crawled_urls_path
            # -----------------------------------------------
            if not output.get("crawled_url"):
                """正常情况下，应该都是没有 crawled_url, 从来都不可能进入到下面的else判断

                因为只有在这里，才会被加上crawled_url，而能进入这里，要么下一步就是成功，要么下一步就是失败。
                """
                output["crawled_url"] = response.url

                meta_easyspider = response.meta.get("easyspider") or {}

                # 写入整个爬虫的路径过程，如果预先就有请求过程，那么就在之前的请求过程后面追加
                crawled_urls_path = meta_easyspider.get("crawled_urls_path") or []
                if not response.meta.get("redirect_urls"):
                    # 如果没有被跳转的情况下，就是当前的链接，加上历史的请求过程
                    crawled_urls_path = crawled_urls_path + [response.url]
                else:
                    # 有跳转，那么跳转的路径，也需要记录
                    crawled_urls_path = crawled_urls_path + response.meta.get("redirect_urls")
                # 回写进去数据结果，保存整个爬虫的经历路径
                output["crawled_urls_path"] = crawled_urls_path
            else:
                """是不可能进入这个判断的 ！！！
                """
                output["crawled_urls_path"] = response.meta.get("easyspider", {}).get("crawled_urls_path") or []
                output["crawled_urls_path"] += [response.url]
            # -----------------------------------------------
            # 重大改变： 添加了response参数，方便回推
            dfd = self.itemproc.process_item(output, spider, response)
            dfd.addBoth(self._itemproc_finished, output, response, spider)
            return dfd
        elif output is None:
            # 强制要求，必须输出结果，哪怕是一个空的 {}
            # 不输出结果的话，就不会加入dupefilter，将会别其他爬虫或者它自己重复性消费
            # 这样做的目的是精细化抓取，避免因为被屏蔽等原因，虽然正常，但是没有获得结果
            # 导致链接丢失，从而丢掉数据
            pass
        else:
            typename = type(output).__name__
            logger.error('Spider must return Request, BaseItem, dict or None, '
                         'got %(typename)r in %(request)s',
                         {'request': request, 'typename': typename},
                         extra={'spider': spider})

    def _itemproc_finished(self, output, item, response, spider):
        """pipeline 全部处理完之后，就该是标记已经完成，即 添加进pipeline
        (只有当全部的流程，全部的pipeline走完，仍然没问题，才能被标记成功)

        ItemProcessor finished for the given ``item`` and returned ``output``
        """
        self.slot.itemproc_size -= 1
        # -------------------------
        rq = response.request
        # --------------------------
        if isinstance(output, Failure) or not rq:
            """另外一种常见的情况就是，在处理pipeline的时候出问题了
            比如说 DataError: (_mysql_exceptions.DataError) (1406, "Data too long for column 'categories' at row 1")
            此时就不应该添加successed_url 而是直接提交到failed_url 去保存来报警
            并且不应该添加 【dupefilter】
            (不添加很难做，暂时添加，后面把 failed_url 重新放回start_urs的时候，给定dont_filter=True来解决这个问题)
            """
            ex = output.value
            if isinstance(ex, DropItem):
                logkws = self.logformatter.dropped(item, ex, response, spider)
                logger.log(*logformatter_adapter(logkws), extra={'spider': spider})
                return self.signals.send_catch_log_deferred(
                    signal=signals.item_dropped, item=item, response=response,
                    spider=spider, exception=output.value)
            else:
                logger.error('Error processing %(item)s', {'item': item},
                             exc_info=failure_to_exc_info(output),
                             extra={'spider': spider})
                # 既然出错，那么就要block_callback记录，重新放回起始队列
                r_copy = response.copy()
                # 如果直接从 yiled Reuest 过来的话，那就是没有带上 easyspider 信息的
                if "easyspider" not in r_copy.request.meta:
                    r_copy.request.meta["easyspider"] = {}
                r_copy.request.meta["easyspider"]["from_retry"] = 1
                # 带上item 方便检查错误在哪
                msg = "Error processing %s ; %s" % (json.dumps(item), output.getTraceback())
                spider.put_back_2_start_url(response,
                                            exc_info=msg,
                                            )

        else:
            logkws = self.logformatter.scraped(output, response, spider)
            logger.log(*logformatter_adapter(logkws), extra={'spider': spider})
            # ----Add fingerprint--------
            fp = request_fingerprint(rq)
            try:
                spider.server.sadd("%s:dupefilter" % spider.name, fp)
            except Exception:
                logger.exception("when _itemproc_finished, add into request fingerprint failed....")
            # ------remove request record beacuse successed-----

            # data = self._encode_request_for_remove_successed_request(response.request, spider)
            # spider.server.execute_command("ZREM", '%(spider)s:requests' % {"spider": spider.name}, data)
            # print "\n\nremove .... remove \n\n"
            # print "\n\n request is data %s" % repr(data)
            # --------------------------------------------------

            # ------add successed_urls urls-----
            # ------2018-03-13 13:17:19  不再保存successed_url 会导致redis过大
            # successed_request = spider.parse_request_to_dict(response)
            # try:
            #     spider.server.sadd("%s:successed_urls" % spider.name, json.dumps(successed_request))
            # except Exception:
            #     logger.exception("in _itemproc_finished, add successed request failed....")
            # --------------------------------------------------
            return self.signals.send_catch_log_deferred(
                signal=signals.item_scraped, item=output, response=response,
                spider=spider)

    def _encode_request_for_remove_successed_request(self, request, spider):
        """Encode a request object"""
        obj = request_to_dict(request, spider)
        return picklecompat.dumps(obj)

    def enqueue_scrape(self, response, request, spider):
        """某些极端的情况下，可能scraper本身出错，这个就严重了....需要汇报作者
        一般情况下来说，作为基础组件的scraper，是绝对不会出错的
        """
        slot = self.slot
        dfd = slot.add_response_request(response, request)

        def finish_scraping(_):
            slot.finish_response(response, request)
            self._check_if_closing(spider, slot)
            self._scrape_next(spider, slot)
            return _
        dfd.addBoth(finish_scraping)

        def scraper_bug(f):
            """出现了严重的问题
            """
            r_copy = response.copy()
            # 如果直接从 yiled Reuest 过来的话，那就是没有带上 easyspider 信息的
            if "easyspider" not in r_copy.request.meta:
                r_copy.request.meta["easyspider"] = {"from_retry": 1}
            else:
                r_copy.request.meta["easyspider"]["from_retry"] += 1

            reason = 'Scraper bug processing %(request)s please contact hhczy1003@163.com for help' % {'request': request}
            logger.error(reason, exc_info=failure_to_exc_info(f), extra={'spider': spider})
            # 很有可能就是调用了 blocked_call_back ，然后blocked_call_back 调用了report 无法输出导致错误的，所以这里就直接提交到redis 去，不再处理
            # spider.blocked_call_back(response, reason=reason, exc_info=f.getTraceback())
            spider.put_back_2_start_url(r_copy,
                                        exc_info=u"%s; %s" % (reason, f.getTraceback()),
                                        )

        dfd.addErrback(scraper_bug)
        self._scrape_next(spider, slot)
        return dfd



# -*- coding: utf-8 -*-
# @Author: hang.zhang
# @Date:   2017-07-17 11:32:39
# @Last Modified by:   zhangTian
# @Last Modified time: 2017-07-18 01:01:08
import sys
sys.path.append("../easyspider/utils")
sys.path.append("../easyspider")

from dbConn_for_etl import MongoConnect, BaseRedis
import argparse
import settings
from tools import get_time
from apscheduler.schedulers.blocking import BlockingScheduler
"""

1. 先取出现在的mongo结果
2. 从redis或者其他存储中取出历史的结果
3. 比较两个结果，新就更新
4. 把比较的结果存入存储中

历史结果存入mysql 待定
"""

class result_monitor_base(object):

    # 解析命令行参数，数据库连接
    def __init__(self):
        self.args = self.parse_args()
        # 连接mongo
        self.mongo_server = self.get_mongo_connect(self.args.mongo_url, self.args.mongo_db_name)
        # 连接redis
        self.redis_server = BaseRedis(self.args.redis_url)
        # 需要关注的collections 集合
        # 操他妈的...集合是会变动的怎么能写死！！弄得还以为是pymongo有缓存了呢
        # self.mongo_collections = self.get_mongo_collections()

    def parse_args(self):
        parser = argparse.ArgumentParser(description="结果监控服务")
        parser.add_argument("-r", "--redis-url", help="redis连接地址，默认从settings文件中读取")
        parser.add_argument("-m", "--mongo-url", help="mongodb连接地址，默认从settings文件中读取")
        parser.add_argument("-d", "--mongo-db-name", help="mongodb的数据库，默认从settings文件中读取")
        parser.add_argument("-c", "--mongo-collection", help="mongodb的集合名，默认为该数据库下的全部监控")
        parser.add_argument("-k", "--redis-key", help="存储历史mongo结果的redis key，默认从settings文件中读取")
        args = parser.parse_args()

        if not args.redis_url:
            args.redis_url = settings.REDIS_URL if hasattr(settings, "REDIS_URL") else "redis://127.0.0.1:6379"

        if not args.mongo_url:
            args.mongo_url = settings.MONGO_URL if hasattr(settings, "MONGO_URL") else "mongodb://127.0.0.1:27017/"

        if not args.mongo_db_name:
            args.mongo_db_name = settings.MONGO_DB_NAME if hasattr(settings, "MONGO_DB_NAME") else "easyspider"

        if not args.redis_key:
            args.redis_key = "easyspider:mongo_result_state"

        return args

    def get_mongo_connect(self, mongo_url, mongo_db_name):
        server = MongoConnect(self.args.mongo_url)
        server.select_db(self.args.mongo_db_name)
        return server

    def get_mongo_collections(self):
        if self.args.mongo_collection:
            return [self.args.mongo_collection]
        return self.mongo_server.list_all_collections()

    def run(self):
        # 集合会变动，每次都要重新查询有多少个集合
        mongo_collections = self.get_mongo_collections()
        # 查询出当前的mongo 结果记录
        current_mongo_result_state = self.get_current_mongo_result_state(mongo_collections)
        # 查询出上一次结果记录(注意如果没有记录上一次变更时间数据的话，那么就以当前的为准)
        last_mongo_result_state = self.get_last_mongo_result_state()
        # 解析当前的mongo进度
        mongo_result_rate = self.parse_mongo_result_rate(current_mongo_result_state, last_mongo_result_state)
        # 将进度存入
        self.dumps_current_mongo_reulst_state(mongo_result_rate)

    def get_current_mongo_result_state(self, mongo_collections):
        info = {}
        for c in mongo_collections:
            current_mongo_result_count = self.mongo_server.collection_count(c)
            info[c] = {
                "result_count": current_mongo_result_count,
                "check_time": get_time()
            }
        return info

    def get_last_mongo_result_state(self):
        return self.redis_server.hgetall(self.args.redis_key)[1]

    def parse_mongo_result_rate(self, current_mongo_result_state, last_mongo_result_state):
        # 不能用老记录去更新。因为在最开始，老记录是没有的
        # 但是使用新记录去更新，又会存在，如果新记录少了删除了，老记录不能被发现的问题
        # 那么就预先检查一次，新纪录没有了，老记录中有的，就置0

        # 老记录检查是否有删除的项目
        for k in last_mongo_result_state.keys():
            if not current_mongo_result_state.get(k):
                last_mongo_result_state[k] = None

        for k, v in current_mongo_result_state.items():
            # 历史记录，因为写入的时候是按照逗号分隔的
            # 第一列代表当前的值，第二列代表当前更新时间，第三列代表上次更新值，第四列代表浮动，第五列代表变更时间
            # 上次计数
            last_count =  last_mongo_result_state.get(k, "0,0,0,0,0").split(",")[0]
            # 上次检查时间
            last_check_time = last_mongo_result_state.get(k, "0,0,0,0,0").split(",")[1]
            # 上次变更时间
            last_modify_time = last_mongo_result_state.get(k, "0,0,0,0,0").split(",")[4]
            # 当前计数
            current_count = v.get("result_count")
            current_check_time = v.get("check_time")
            # 和历史记录比较的浮动 当前数目 - 上次存储的当前数目
            increase_count = int(current_count) - int(last_count)
            # 如果为0, 说明上次的记录没有变化，变更时间是上次的，不然的话，变更时间就是现在，当前变更
            modify_time = last_modify_time if increase_count == 0 else current_check_time

            current_result_state = "%s,%s,%s,%s,%s" % (current_count, current_check_time, last_count, increase_count, modify_time)
            last_mongo_result_state[k] = current_result_state
        return last_mongo_result_state


    def dumps_current_mongo_reulst_state(self, mongo_result_rate):
        for collection, values in mongo_result_rate.items():
            # 为空说明这里的集合已经在mongo中被删除
            if not values:
                # print "del %s" % collection
                self.redis_server.hdel(self.args.redis_key, collection)
            else:
                # print "update %s" % collection
                self.redis_server.hset(self.args.redis_key, collection, values)


if __name__ == '__main__':
    server = result_monitor_base()
    # server.run()
    scheduler = BlockingScheduler(daemonic=True)
    scheduler.add_job(server.run, 'interval', seconds=20)
    scheduler.start()
    # import time
    # while True:
    #     server.run()
    #     time.sleep(10)
    #     print "-------"




# -*- coding: utf-8 -*-
# @Author: zhangTian
# @Email:  hhczy1003@163.com
# @Date:   2017-07-08 13:48:45
# @Last Modified by:   hang.zhang
# @Last Modified time: 2018-03-18 06:57:22


# TODO: 队列监控反馈
import logging
from scrapy_redis.scheduler import Scheduler
from twisted.internet.threads import deferToThread
logger = logging.getLogger(__name__)


class easyScheduler(Scheduler):

    def has_pending_requests(self):
        """in this method, you should only return True regardless what len(self) is

        because in a extreme case, spider machine may lost connection with the redis
        (ADSL dailing interval or even worse case: the ADSL interface had crush down and must reboot to resume),

        if can not connect to redis, len(self) will casuse Exception, because of the Exception, crawler can not
        run into next_request, and run into a useless loop, never require more request from redis anymore(because next_requsts not called)
        """
        try:
            return len(self) > 0
        except Exception:
            logger.exception("in easyScheduler: check has_pending_requests,  len(self) failed, return false")
            return False

    def enqueue_request(self, request):
        if not request.dont_filter and self.df.request_seen(request):
            self.df.log(request, self.spider)
            return False
        if self.stats:
            self.stats.inc_value('scheduler/enqueued/redis', spider=self.spider)

        return deferToThread(self.defer_push, request)

    def defer_push(self, request):
        # loop until connected to redis
        while True:
            try:
                self.queue.push(request)
                return True
            except Exception:
                logger.exception("in easyScheduler: enqueue_request,  self.queue.push(request) failed, return True %s" % request)
                # 2018-03-18 06:19:52： 暂时想不出好办法，在 掉线的时候怎么办。。。
                # 出现的典型状况是，adsl 掉线的时候，刚好运行到这里，然后就一直卡主，在这里循环，执行权限没有放开
                # 想搞个子进程通信，把adsl独立开，这样即使这里一直被卡住，adsl 到时间了，也能自动恢复，等它恢复了，这里也就好了
                # return True
